from datetime import timedelta
from typing import Optional

import pandas as pd
import plotly.express as px
from matplotlib.figure import Figure

from sampo.schemas import Time
from sampo.utilities.visualization.base import VisualizationMode, visualize


def schedule_gant_chart_fig(schedule_dataframe: pd.DataFrame,
                            visualization: VisualizationMode,
                            remove_service_tasks: bool = False,
                            fig_file_name: Optional[str] = None,
                            color_type: str = 'contractor') -> Figure | None:
    """
    Creates and saves a gant chart of the scheduled tasks to the specified path.

    :param fig_file_name:
    :param visualization:
    :param remove_service_tasks:
    :param schedule_dataframe: Pandas DataFrame with the information about schedule
    :param color_type defines what tasks color means
    """
    if remove_service_tasks:
        schedule_dataframe = schedule_dataframe.loc[~schedule_dataframe.loc[:, 'task_name'].str.contains('марке')]
        schedule_dataframe = schedule_dataframe.loc[schedule_dataframe.loc[:, 'volume'] >= 0.1]

    schedule_dataframe = schedule_dataframe.rename({'workers': 'workers_dict'}, axis=1)
    schedule_dataframe.loc[:, 'workers'] = schedule_dataframe.loc[:, 'workers_dict']\
        .apply(lambda x: x.replace(", '", ", <br>'"))

    schedule_start = schedule_dataframe.loc[:, 'start'].min()
    schedule_finish = schedule_dataframe.loc[:, 'finish'].max()
    visualization_start_delta = timedelta(days=2)
    visualization_finish_delta = timedelta(days=(schedule_finish - schedule_start).days // 3)

    def create_delivery_row(i, mat_name, material) -> dict:
        return {'idx': i,
                'contractor': material[-1],
                'cost': 0,
                'volume': material[0],
                'duration': 0,
                'measurement': 'unit',
                'workers_dict': '',
                'workers': '',
                'task_name_mapped': mat_name,
                'task_name': '',
                'zone_information': '',
                'start': timedelta(material[1].value) + schedule_start,
                'finish': timedelta(material[2].value) + schedule_start}

    sworks = schedule_dataframe['scheduled_work_object'].copy()
    idx = schedule_dataframe['idx'].copy()

    def get_delivery_info(swork) -> str:
        return '<br>' + '<br>'.join([f'{mat[0][0]}: {mat[0][1]}' for mat in swork.materials.delivery.values()])

    schedule_dataframe['material_information'] = sworks.apply(get_delivery_info)

    mat_delivery_row = []

    # create material delivery information
    for i, swork in zip(idx, sworks):
        delivery = swork.materials.delivery
        for name, mat_info in delivery.items():
            if delivery:
                mat_delivery_row.append(create_delivery_row(i, name, mat_info[0]))

    def create_zone_row(i, zone_names, zone) -> dict:
        return {'idx': i,
                'contractor': 'Access cards',
                'cost': 0,
                'volume': 0,
                'duration': 0,
                'measurement': 'unit',
                'workers_dict': '',
                'workers': '',
                'task_name_mapped': zone_names,
                'task_name': '',
                'zone_information': '',
                'start': timedelta(int(zone.start_time)) + schedule_start,
                'finish': timedelta(int(zone.end_time)) + schedule_start}

    sworks = schedule_dataframe['scheduled_work_object'].copy()
    idx = schedule_dataframe['idx'].copy()

    def get_zone_usage_info(swork) -> str:
        return '<br>' + '<br>'.join([f'{zone.name}: {zone.to_status}' for zone in swork.zones_pre])

    schedule_dataframe['zone_information'] = sworks.apply(get_zone_usage_info)

    access_cards = []

    # create zone information
    for i, swork in zip(idx, sworks):
        zone_names = '<br>' + '<br>'.join([zone.name for zone in swork.zones_pre])
        for zone in swork.zones_pre:
            access_cards.append(create_zone_row(i, zone_names, zone))
        zone_names = '<br>' + '<br>'.join([zone.name for zone in swork.zones_post])
        for zone in swork.zones_post:
            access_cards.append(create_zone_row(i, zone_names, zone))

    schedule_dataframe = pd.concat([schedule_dataframe, pd.DataFrame.from_records(access_cards), pd.DataFrame.from_records(mat_delivery_row)])

    if color_type == 'contractor':
        schedule_dataframe['color'] = schedule_dataframe[['task_name', 'contractor']] \
            .apply(lambda r: 'Defect' if ':' in r['task_name'] else r['contractor'], axis=1)
    elif color_type == 'priority':
        schedule_dataframe['color'] = schedule_dataframe['scheduled_work_object'].apply(lambda x: f'Priority {x.priority}')

    schedule_dataframe['idx'] = (schedule_dataframe[['idx', 'task_name']]
                                 .apply(lambda r: schedule_dataframe[schedule_dataframe['task_name'] ==
                                                                     r['task_name'].split('&')[0]]['idx'].iloc[0]
                                 if ':' in r['task_name'] else r['idx'], axis=1))

    fig = px.timeline(schedule_dataframe, x_start='start', x_end='finish', y='idx', hover_name='task_name',
                      color=schedule_dataframe.loc[:, 'color'] if 'color' in schedule_dataframe.columns else None,
                      hover_data={'start': True,
                                  'finish': True,
                                  'task_name_mapped': True,
                                  'cost': True,
                                  'volume': True,
                                  'measurement': True,
                                  'workers': True,
                                  'zone_information': True,
                                  'material_information': True},
                      title=f"{'Project tasks - Gant chart'}",
                      category_orders={'idx': list(schedule_dataframe.idx)},
                      text='task_name')

    fig.update_traces(textposition='outside')

    fig.update_yaxes(showticklabels=False, title_text='Project tasks',
                     type='category')

    fig.update_xaxes(type='date',
                     range=[schedule_start - visualization_start_delta,
                            schedule_finish + visualization_finish_delta],
                     title_text='Date')

    fig.update_layout(autosize=True, font_size=12)
    # fig.update_layout(height=1000)

    return visualize(fig, mode=visualization, file_name=fig_file_name)



from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.schedule import Schedule
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.utilities.serializers import custom_serializer


class ScheduledProject(AutoJSONSerializable['ScheduledProject']):

    ignored_fields = ['raw_schedule', 'raw_wg']

    def __init__(self, wg: WorkGraph, raw_wg: WorkGraph, contractors: list[Contractor], schedule: Schedule):
        """
        Contains schedule and all information about its creation
        :param wg: the original work graph
        :param raw_wg: restructured work graph, which was given directly to scheduler to produce this schedule
        :param contractors: list of contractors
        :param schedule: the raw schedule received directly from scheduler
        """
        # the final variant of schedule, without any technical issues
        self.schedule = schedule.unite_stages()
        # the raw schedule, with inseparables
        self.raw_schedule = schedule
        # the original work graph
        self.wg = wg
        # internally processed work graph, with inseparables
        self.raw_wg = raw_wg
        self.contractors = contractors

    @custom_serializer('contractors')
    def serialize_contractors(self, value) -> list:
        return [v._serialize() for v in value]

    @classmethod
    @custom_serializer('contractors', deserializer=True)
    def deserialize_equipment(cls, value) -> list[Contractor]:
        return [Contractor._deserialize(v) for v in value]



from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import numpy as np

from sampo.schemas.time import Time
from sampo.schemas.serializable import AutoJSONSerializable


@dataclass
class Zone:
    name: str
    status: int


class ZoneStatuses(ABC):
    @abstractmethod
    def statuses_available(self) -> int:
        """
        :return: number of statuses available
        """
        ...

    @abstractmethod
    def match_status(self, target: int, to_compare: int) -> bool:
        """
        :param target: statuses that should match
        :param to_compare: status that should be matched
        :return: does target match to_compare
        """
        ...


class DefaultZoneStatuses(ZoneStatuses):
    """
    Statuses: 0 - not stated, 1 - opened, 2 - closed
    """

    def statuses_available(self) -> int:
        return 3

    def match_status(self, status_to_check: int, required_status: int) -> bool:
        return required_status == 0 or status_to_check == required_status


@dataclass
class ZoneConfiguration:
    start_statuses: dict[str, int] = field(default_factory=dict)
    end_statuses: dict[str, int] = field(default_factory=dict)
    time_costs: np.ndarray = field(default_factory=lambda: np.array([[]]))
    statuses: ZoneStatuses = field(default_factory=lambda: DefaultZoneStatuses())

    def change_cost(self, from_status: int, to_status: int) -> int:
        return self.time_costs[from_status, to_status]


@dataclass
class ZoneTransition(AutoJSONSerializable['ZoneTransition']):
    name: str
    from_status: int
    to_status: int
    start_time: Time
    end_time: Time

    def __str__(self) -> str:
        return f'Access card {self.name} status {self.from_status} -> {self.to_status}'



from uuid import uuid4

import numpy as np

from sampo.schemas import GraphNode, Worker, WorkTimeEstimator
from sampo.scheduler.utils.time_computaion import calculate_working_time_cascade


def get_chain_duration(node: GraphNode, assigned_workers_amounts: np.ndarray, work_estimator: WorkTimeEstimator) -> int:
    work_unit = node.work_unit

    passed_workers = [Worker(str(uuid4()), req.kind, assigned_amount)
                      for req, assigned_amount in zip(work_unit.worker_reqs, assigned_workers_amounts)]

    chain_duration = calculate_working_time_cascade(node, passed_workers, work_estimator).value + 1

    return chain_duration



import pandas as pd

from sampo.schemas import WorkGraph, Contractor, WorkTimeEstimator
from sampo.schemas.structure_estimator import StructureEstimator
from sampo.schemas.time_estimator import DefaultWorkEstimator
from sampo.userinput import CSVParser
from sampo.utilities.name_mapper import NameMapper


class PreparationPipeline:

    def __init__(self,
                 wg: str | pd.DataFrame | WorkGraph = None,
                 history: str | pd.DataFrame | None = None,
                 contractors: str | pd.DataFrame | list[Contractor] | None = None):
        self._wg = wg
        self._contractors = contractors
        self._history = history if history is not None else pd.DataFrame(columns=['marker_for_glue', 'work_name',
                                                                                  'first_day', 'last_day',
                                                                                  'upper_works', 'work_name_clear_old',
                                                                                  'smr_name', 'work_name_clear',
                                                                                  'granular_smr_name'])
        self._fix_edges_with_history = False
        self._fill_edges_with_history = False
        self._sep_wg = ';'
        self._sep_history = ';'
        self._name_mapper = None
        self._work_estimator = DefaultWorkEstimator()
        self._structure_estimator = None

    def wg(self, wg: str | pd.DataFrame | WorkGraph):
        self._wg = wg
        return self

    def history(self, history: str | pd.DataFrame):
        self._history = history
        return self

    def contractors(self, contractors: str | pd.DataFrame | list[Contractor]):
        self._contractors = contractors
        return self

    def fix_edges_with_history(self) -> 'PreparationPipeline':
        self._fix_edges_with_history = True
        return self

    def fill_edges_with_history(self) -> 'PreparationPipeline':
        self._fill_edges_with_history = True
        return self

    def sep_wg(self, sep: str) -> 'PreparationPipeline':
        self._sep_wg = sep
        return self

    def sep_history(self, sep: str) -> 'PreparationPipeline':
        self._sep_history = sep
        return self

    def name_mapper(self, name_mapper: NameMapper) -> 'PreparationPipeline':
        self._name_mapper = name_mapper
        return self

    def work_estimator(self, work_estimator: WorkTimeEstimator) -> 'PreparationPipeline':
        self._work_estimator = work_estimator
        return self

    def structure_estimator(self, structure_estimator: StructureEstimator) -> 'PreparationPipeline':
        self._structure_estimator = structure_estimator
        return self

    def _prepare_works_info(self) -> pd.DataFrame:
        wg = self._wg.to_frame(save_req=True) if isinstance(self._wg, WorkGraph) else self._wg
        return CSVParser.read_graph_info(project_info=wg,
                                         history_data=self._history,
                                         sep_wg=self._sep_wg,
                                         sep_history=self._sep_history,
                                         name_mapper=self._name_mapper,
                                         all_connections=not self._fix_edges_with_history and
                                                         not self._fill_edges_with_history,
                                         change_connections_info=self._fix_edges_with_history)

    def build_wg(self) -> WorkGraph:
        wg = CSVParser.work_graph(
            works_info=self._prepare_works_info(),
            name_mapper=self._name_mapper,
            work_resource_estimator=self._work_estimator
        )

        if self._structure_estimator is not None:
            wg = self._structure_estimator.restruct(wg)

        return wg

    def build_wg_and_contractors(self) -> tuple[WorkGraph, list[Contractor]]:
        return CSVParser.work_graph_and_contractors(
            works_info=self._prepare_works_info(),
            name_mapper=self._name_mapper,
            contractor_info=self._contractors,
            work_resource_estimator=self._work_estimator
        )



from copy import deepcopy

from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.utils import uuid_str
from sampo.schemas.works import WorkUnit


def copy_graph_node(node: GraphNode, new_id: int | str | None = None,
                    change_id: bool = True) -> tuple[GraphNode, tuple[str, str]]:
    """
    Makes a deep copy of GraphNode without edges. It's id can be changed to a new randomly generated or specified one
    :param node: original GraphNode
    :param new_id: specified new id
    :param change_id: do ids in the new graph need to be changed
    :return: copy of GraphNode and pair(old node id, new node id)
    """
    if change_id:
        new_id = new_id or uuid_str()
        new_id = str(new_id) if isinstance(new_id, int) else new_id
    else:
        new_id = node.work_unit.id
    wu = node.work_unit
    new_wu = WorkUnit(id=new_id, name=wu.name,
                      worker_reqs=deepcopy(wu.worker_reqs),
                      material_reqs=deepcopy(wu.material_reqs),
                      equipment_reqs=deepcopy(wu.equipment_reqs),
                      object_reqs=deepcopy(wu.object_reqs),
                      zone_reqs=deepcopy(wu.zone_reqs),
                      group=wu.group,
                      is_service_unit=wu.is_service_unit, volume=wu.volume, volume_type=wu.volume_type)
    return GraphNode(new_wu, []), (wu.id, new_id)


def restore_parents(new_nodes: dict[str, GraphNode], original_wg: WorkGraph, old_to_new_ids: dict[str, str],
                    excluded_ids: set[str]) -> None:
    """
    Restores edges in GraphNode for copied WorkGraph with changed ids
    :param new_nodes: needed copied nodes
    :param original_wg: original WorkGraph for edge restoring for new nodes
    :param excluded_ids: dictionary of relationships between old ids and new ids
    :param old_to_new_ids: a dictionary linking the ids of GraphNodes of the original graph and the new GraphNode ids
    :return:
    """
    for node in original_wg.nodes:
        if node.id in old_to_new_ids and node.id not in excluded_ids:
            new_node = new_nodes[old_to_new_ids[node.id]]
            new_node.add_parents([(new_nodes[old_to_new_ids[edge.start.id]], edge.lag, edge.type)
                                  for edge in node.edges_to
                                  if edge.start.id in old_to_new_ids and edge.start.id not in excluded_ids])


def prepare_work_graph_copy(wg: WorkGraph, excluded_nodes: list[GraphNode] = [], use_ids_simplification: bool = False,
                            id_offset: int = 0, change_id: bool = True) -> tuple[dict[str, GraphNode], dict[str, str]]:
    """
    Makes a deep copy of the GraphNodes of the original graph with new ids and updated edges,
    ignores all GraphNodes specified in the exception list and GraphEdges associated with them
    :param wg: original WorkGraph for copy
    :param excluded_nodes: GraphNodes to be excluded from the graph
    :param use_ids_simplification: If true, creates short numeric ids converted to strings,
    otherwise uses uuid to generate id
    :param id_offset: Shift for numeric ids, used only if param use_ids_simplification is True
    :param change_id: Do IDs in the new graph need to be changed
    :return: A dictionary with GraphNodes by their id
    and a dictionary linking the ids of GraphNodes of the original graph and the new GraphNode ids
    """
    excluded_nodes = {node.id for node in excluded_nodes}
    node_list = [(id_offset + ind, node) for ind, node in enumerate(wg.nodes)] \
        if use_ids_simplification \
        else [(None, node) for node in wg.nodes]
    nodes, old_to_new_ids = list(zip(*[copy_graph_node(node, ind, change_id) for ind, node in node_list
                                       if node.id not in excluded_nodes]))
    id_old_to_new = dict(old_to_new_ids)
    nodes = {node.id: node for node in nodes}
    restore_parents(nodes, wg, id_old_to_new, excluded_nodes)
    return nodes, id_old_to_new


def new_start_finish(original_wg: WorkGraph, copied_nodes: dict[str, GraphNode],
                     old_to_new_ids: dict[str, str]) -> tuple[GraphNode, GraphNode]:
    """
    Prepares new start and finish to create WorkGraph after copying it
    :param original_wg: WorkGraph, on which base prepare_work_graph_copy was run
    :param copied_nodes: New nodes, on which to create new WorkGraph
    :param old_to_new_ids: Dictionary to translate old nodes to new, using their IDs
    :return: new start and new finish nodes, on the base of which to create a WorkGraph
    """
    new_start = copied_nodes[old_to_new_ids[original_wg.start.id]]
    new_finish = copied_nodes[old_to_new_ids[original_wg.finish.id]]
    return new_start, new_finish



import logging

from sampo.backend.default import DefaultComputationalBackend

logging.basicConfig(format='[%(name)s] [%(levelname)s] %(message)s', level=logging.INFO)


class SAMPO:
    backend = DefaultComputationalBackend()
    logger = logging.getLogger('SAMPO')



from abc import ABC, abstractmethod
from copy import deepcopy
from random import Random
from typing import Callable

from sampo.schemas.graph import WorkGraph, GraphNode


class Obstruction(ABC):
    """
    Tests the probability and, if it's true, apply the obstruction.
    """
    def __init__(self, probability: float, rand: Random):
        self._probability = probability
        self._rand = rand

    def generate(self, wg: WorkGraph):
        """
        Tests the probability and, if it's true, apply the obstruction.

        :param wg: given WorkGraph
        """
        if self._rand.random() < self._probability:
            self.apply(wg)

    @abstractmethod
    def apply(self, wg: WorkGraph):
        """
        The main method of the obstruction.

        Should apply provided obstruction to the given WorkGraph in-place.

        :param wg: given WorkGraph
        """
        ...


class OneInsertObstruction(Obstruction):
    """
    Applying seeks the random part of given WorkGraph and inserts it into that point.
    """

    def __init__(self, probability: float, rand: Random, insert_wg_getter: Callable[[Random], WorkGraph]):
        super().__init__(probability, rand)
        self._insert_wg_getter = insert_wg_getter

    @staticmethod
    def from_static_graph(probability: float, rand: Random, insert_wg: WorkGraph) -> 'OneInsertObstruction':
        return OneInsertObstruction(probability, rand, lambda _: deepcopy(insert_wg))

    def apply(self, wg: WorkGraph):
        # get the insert graph
        insert_wg = self._insert_wg_getter(self._rand)
        # get the insert point
        insert_node: GraphNode = self._rand.sample(wg.nodes, 1)[0]
        # insert
        insert_wg.start.add_parents([insert_node])

        if insert_node.children:
            # get the insert end point
            insert_node_child: GraphNode = self._rand.sample(insert_node.children, 1)[0]
            # insert
            insert_node_child.add_parents([insert_wg.start])

        wg.reinit()




from collections import defaultdict
from typing import Iterable

from toposort import toposort_flatten

from sampo.schemas import Worker, Contractor, WorkGraph, GraphNode
from sampo.schemas.types import WorkerName, ContractorName

WorkerContractorPool = dict[WorkerName, dict[ContractorName, Worker]]


def get_worker_contractor_pool(contractors: Iterable[Contractor]) -> WorkerContractorPool:
    """
    Gets worker-contractor dictionary from contractors list.
    Alias for frequently used functionality.

    :param contractors: list of all the considered contractors
    :return: dictionary of workers by worker name, next by contractor id
    """
    worker_pool = defaultdict(dict)
    for contractor in contractors:
        for name, worker in contractor.workers.items():
            worker_pool[name][contractor.id] = worker.copy()
    return worker_pool


def get_head_nodes_with_connections_mappings(wg: WorkGraph
                                             ) -> tuple[list[GraphNode], dict[str, set[str]], dict[str, set[str]]]:
    """
    Identifies 'head nodes' in a WorkGraph and reconstructs their inter-node dependencies.

    Head nodes are defined as the first nodes of inseparable chains or standalone nodes
    that are not part of an inseparable chain (i.e., they are not 'inseparable sons').
    This function effectively flattens the graph by treating inseparable chains as
    single logical entities represented by their head node, and then re-establishes
    parent-child relationships between these head nodes.

    Args:
        wg: The `WorkGraph` to analyze.

    Returns:
        A tuple containing:
            - A list of `GraphNode` objects representing the head nodes,
              sorted in topological order based on their reconstructed dependencies.
            - A dictionary mapping the ID of each head node to a set of IDs of
              its new 'parent' head nodes. These represent external dependencies
              where a parent of any node within the current head node's inseparable
              chain belongs to another head node's chain.
            - A dictionary mapping the ID of each head node to a set of IDs of
              its new 'child' head nodes. Similar to parents, these represent
              external dependencies where a child of any node within the current
              head node's inseparable chain belongs to another head node's chain.
    """
    # Filter the work graph nodes to identify all 'head nodes'.
    # A head node is one that is not an 'inseparable son', meaning it's either
    # the start of an inseparable chain or a standalone node.
    nodes = [node for node in wg.nodes if not node.is_inseparable_son()]

    # Construct a mapping from any node within an inseparable chain to its
    # corresponding head node.
    node2inseparable_parent = {child: node for node in nodes for child in node.get_inseparable_chain_with_self()}

    # Reconstruct parent-child relationships between the identified head nodes.
    # For each head node, gather all direct parents of any node within its
    # inseparable chain. Then, map these direct parents back to their
    # respective head nodes using `node2inseparable_parent`.
    node_id2parent_ids = {node.id: set(
        node2inseparable_parent[parent].id  # Get the head node ID for the actual parent
        for inseparable in node.get_inseparable_chain_with_self()  # Iterate through all parts of head node's chain
        for parent in inseparable.parents_set  # Get direct parents of each part
    ) - {node.id}  # Remove self-references
                          for node in nodes}

    # Reconstruct child-parent relationships in the same manner as parent-child.
    # For each head node, gather all direct children of any node within its
    # inseparable chain. Then, map these direct children back to their
    # respective head nodes using `node2inseparable_parent`.
    node_id2child_ids = {node.id: set(
        node2inseparable_parent[child].id  # Get the head node ID for the actual child
        for inseparable in node.get_inseparable_chain_with_self()  # Iterate through all parts of head node's chain
        for child in inseparable.children_set  # Get direct children of each part
    ) - {node.id}  # Remove self-references
                         for node in nodes}

    # Perform a topological sort on the head nodes based on their new parent dependencies.
    # This ensures the returned list of head nodes is ordered correctly for scheduling or analysis.
    tsorted_nodes_ids = toposort_flatten(node_id2parent_ids, sort=True)

    # Map the sorted head node IDs back to their corresponding GraphNode objects.
    tsorted_nodes = [wg[node_id] for node_id in tsorted_nodes_ids]

    return tsorted_nodes, node_id2parent_ids, node_id2child_ids



from typing import Type, Callable, Iterable

from sampo.scheduler.base import Scheduler, SchedulerType
from sampo.scheduler.utils import (WorkerContractorPool, get_worker_contractor_pool,
                                   get_head_nodes_with_connections_mappings)
from sampo.scheduler.utils.time_computaion import calculate_working_time_cascade
from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.utils.multi_contractor import run_contractor_search, get_worker_borders
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.resources import Worker
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec, WorkSpec
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator
from sampo.utilities.validation import validate_schedule


# TODO Кажется, это не работает - лаги не учитываются
def get_finish_time_default(node, worker_team, node2swork, spec, assigned_parent_time, timeline,
                            work_estimator) -> Time:
    return timeline.find_min_start_time(node, worker_team, node2swork, spec,
                                        assigned_parent_time, work_estimator) \
        + calculate_working_time_cascade(node, worker_team,
                                         work_estimator)  # TODO Кажется, это не работает - лаги не учитываются


PRIORITIZATION_F = Callable[
    [list[GraphNode], dict[str, set[str]], dict[str, set[str]], WorkTimeEstimator], list[GraphNode]
]
RESOURCE_OPTIMIZE_F = Callable[[GraphNode, list[Contractor], WorkSpec, WorkerContractorPool,
                                dict[GraphNode, ScheduledWork], Time, Timeline, WorkTimeEstimator],
                               tuple[Time, Time, Contractor, list[Worker]]]


class GenericScheduler(Scheduler):
    """
    Implementation of a universal scheme of scheduler.
    It's parametrized by prioritization function and optimization resource function.
    It constructs the end schedulers.
    """

    def __init__(self,
                 scheduler_type: SchedulerType,
                 resource_optimizer: ResourceOptimizer,
                 timeline_type: Type,
                 prioritization_f: PRIORITIZATION_F,
                 optimize_resources_f: RESOURCE_OPTIMIZE_F,
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator()):
        super().__init__(scheduler_type, resource_optimizer, work_estimator)
        self._timeline_type = timeline_type
        self.prioritization = prioritization_f
        self.optimize_resources = optimize_resources_f

    def get_default_res_opt_function(self, get_finish_time=get_finish_time_default) \
            -> Callable[
                [GraphNode, list[Contractor], WorkSpec, WorkerContractorPool,
                 dict[GraphNode, ScheduledWork], Time, Timeline, WorkTimeEstimator],
                tuple[Time, Time, Contractor, list[Worker]]
            ]:
        """
        Here is default resource optimization getter function.

        Constructs function that receives node with necessary scheduling inner info and
        returns start time, finish time and assigned workers for that node.

        :param get_finish_time: function that
        :return: resource optimization function that can be passed as argument when constructing `GenericScheduler`.
        """

        def optimize_resources_def(node: GraphNode, contractors: list[Contractor], spec: WorkSpec,
                                   worker_pool: WorkerContractorPool, node2swork: dict[GraphNode, ScheduledWork],
                                   assigned_parent_time: Time, timeline: Timeline, work_estimator: WorkTimeEstimator):
            def ft_getter(worker_team) -> Time:
                return get_finish_time(node, worker_team, node2swork, spec,
                                       assigned_parent_time, timeline, work_estimator)

            def run_with_contractor(contractor: Contractor) -> tuple[Time, Time, list[Worker]]:
                min_count_worker_team, max_count_worker_team, workers \
                    = get_worker_borders(worker_pool, contractor, node.work_unit.worker_reqs)

                if len(workers) != len(node.work_unit.worker_reqs):
                    return assigned_parent_time, Time.inf(), []

                workers = [worker.copy() for worker in workers]

                # apply worker team spec
                self.optimize_resources_using_spec(node.work_unit, workers, spec,
                                                   lambda optimize_array: self.resource_optimizer.optimize_resources(
                                                       worker_pool, workers,
                                                       optimize_array,
                                                       min_count_worker_team, max_count_worker_team,
                                                       ft_getter))

                c_st, c_ft, _ = timeline.find_min_start_time_with_additional(node, workers, node2swork, spec, None,
                                                                             assigned_parent_time, work_estimator)
                return c_st, c_ft, workers

            return run_contractor_search(contractors, run_with_contractor)

        return optimize_resources_def

    def schedule_with_cache(self,
                            wg: WorkGraph,
                            contractors: list[Contractor],
                            spec: ScheduleSpec = ScheduleSpec(),
                            validate: bool = False,
                            assigned_parent_time: Time = Time(0),
                            timeline: Timeline | None = None,
                            landscape: LandscapeConfiguration() = LandscapeConfiguration()) \
            -> list[tuple[Schedule, Time, Timeline, list[GraphNode]]]:
        # get head nodes with mappings
        head_nodes, node_id2parent_ids, node_id2child_ids = get_head_nodes_with_connections_mappings(wg)

        ordered_nodes = self.prioritization(head_nodes, node_id2parent_ids, node_id2child_ids, self.work_estimator)

        schedule, schedule_start_time, timeline = \
            self.build_scheduler(ordered_nodes, contractors, landscape, spec, self.work_estimator,
                                 assigned_parent_time, timeline)
        schedule = Schedule.from_scheduled_works(
            schedule,
            wg
        )

        if validate:
            validate_schedule(schedule, wg, contractors)

        return [(schedule, schedule_start_time, timeline, ordered_nodes)]

    def build_scheduler(self,
                        ordered_nodes: list[GraphNode],
                        contractors: list[Contractor],
                        landscape: LandscapeConfiguration = LandscapeConfiguration(),
                        spec: ScheduleSpec = ScheduleSpec(),
                        work_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                        assigned_parent_time: Time = Time(0),
                        timeline: Timeline | None = None) \
            -> tuple[Iterable[ScheduledWork], Time, Timeline]:
        """
        Find optimal number of workers who ensure the nearest finish time.
        Finish time is combination of two dependencies: max finish time, max time of waiting of needed workers
        This is selected by iteration from minimum possible numbers of workers until then the finish time is decreasing

        :param landscape:
        :param contractors:
        :param spec: spec for current scheduling
        :param ordered_nodes:
        :param timeline: the previous used timeline can be specified to handle previously scheduled works
        :param assigned_parent_time: start time of the whole schedule(time shift)
        :param work_estimator:
        :return:
        """
        worker_pool = get_worker_contractor_pool(contractors)
        # dict for writing parameters of completed_jobs
        node2swork: dict[GraphNode, ScheduledWork] = {}
        # list for support the queue of workers
        if not isinstance(timeline, self._timeline_type):
            timeline = self._timeline_type(worker_pool, landscape)

        for index, node in enumerate(ordered_nodes):
            work_unit = node.work_unit
            work_spec = spec.get_work_spec(work_unit.id)

            start_time, finish_time, contractor, best_worker_team = self.optimize_resources(node, contractors,
                                                                                            work_spec, worker_pool,
                                                                                            node2swork,
                                                                                            assigned_parent_time,
                                                                                            timeline, work_estimator)

            # we are scheduling the work `start of the project`
            if index == 0:
                # this work should always have start_time = 0, so we just re-assign it
                start_time = assigned_parent_time
                finish_time += start_time

            if index == len(ordered_nodes) - 1:  # we are scheduling the work `end of the project`
                finish_time, finalizing_zones = timeline.zone_timeline.finish_statuses()
                start_time = max(start_time, finish_time)

            # apply work to scheduling
            timeline.schedule(node, node2swork, best_worker_team, contractor, work_spec,
                              start_time, work_spec.assigned_time, assigned_parent_time, work_estimator)

            if index == len(ordered_nodes) - 1:  # we are scheduling the work `end of the project`
                node2swork[node].zones_pre = finalizing_zones

        return node2swork.values(), assigned_parent_time, timeline



from typing import Callable
from itertools import chain

import numpy as np
import random

from sampo.utilities.priority import extract_priority_groups_from_nodes
from sampo.schemas.graph import GraphNode


def map_lft_lst(head_nodes: list[GraphNode],
                node_id2child_ids: dict[str, set[str]],
                node_id2duration: dict[str, int]) -> tuple[dict[str, int], dict[str, int]]:
    project_max_duration = sum(node_id2duration.values())
    node_id2lft = {head_nodes[-1].id: project_max_duration}
    node_id2lst = {head_nodes[-1].id: project_max_duration - node_id2duration[head_nodes[-1].id]}

    for node in reversed(head_nodes[:-1]):
        suc_lst = [node_id2lst[child_id] for child_id in node_id2child_ids[node.id] if child_id in node_id2lst]
        node_id2lft[node.id] = min(suc_lst, default=0)
        node_id2lst[node.id] = node_id2lft[node.id] - node_id2duration[node.id]

    return node_id2lft, node_id2lst


def lft_prioritization(head_nodes: list[GraphNode],
                       node_id2parent_ids: dict[str, set[str]],
                       node_id2child_ids: dict[str, set[str]],
                       node_id2duration: dict[str, int],
                       core_f: Callable[[list[GraphNode],
                                         dict[str, set[str]],
                                         dict[str, set[str]],
                                         dict[str, int],
                                         list[list[GraphNode]],
                                         random.Random], list[GraphNode]],
                       rand: random.Random() = random.Random()) -> list[GraphNode]:
    """
    Return ordered critical nodes.
    Finish time is depended on these ordered nodes.
    """

    # form groups
    groups = extract_priority_groups_from_nodes(head_nodes)
    groups = [groups[priority] for priority in sorted(groups.keys())]

    result = core_f(head_nodes, node_id2parent_ids, node_id2child_ids, node_id2duration, groups, rand)

    return result


def lft_prioritization_core(head_nodes: list[GraphNode],
                            node_id2parent_ids: dict[str, set[str]],
                            node_id2child_ids: dict[str, set[str]],
                            node_id2duration: dict[str, int],
                            groups: list[list[GraphNode]],
                            rand: random.Random = random.Random()) -> list[GraphNode]:
    """
    Return nodes ordered by MIN-LFT priority rule.
    """
    node_id2lft, _ = map_lft_lst(head_nodes, node_id2child_ids, node_id2duration)

    ordered_groups_it = (sorted(group, key=lambda node: node_id2lft[node.id]) for group in groups)
    ordered_nodes = list(chain.from_iterable(ordered_groups_it))

    return ordered_nodes


def lft_randomized_prioritization_core(head_nodes: list[GraphNode],
                                       node_id2parent_ids: dict[str, set[str]],
                                       node_id2child_ids: dict[str, set[str]],
                                       node_id2duration: dict[str, int],
                                       groups: list[list[GraphNode]],
                                       rand: random.Random = random.Random()) -> list[GraphNode]:
    """
    Return ordered nodes sampled by MIN-LFT and MIN-LST priority rules.
    """
    node_id2group_priority = {node.id: i for i, group in enumerate(groups) for node in group}

    def is_eligible(node_id):
        return node_id2parent_ids[node_id].issubset(selected_ids_set)

    nodes2lft, nodes2lst = map_lft_lst(head_nodes, node_id2child_ids, node_id2duration)

    ordered_node_ids = []
    selected_ids_set = set()
    candidates = {head_nodes[0].id}

    while candidates:
        eligibles = [node_id for node_id in candidates if is_eligible(node_id)]
        min_group_priority = min(node_id2group_priority[node_id] for node_id in eligibles)
        eligibles = [node_id for node_id in eligibles if node_id2group_priority[node_id] == min_group_priority]

        priority_mapper = nodes2lft if rand.random() < 0.5 else nodes2lst

        weights = np.array([priority_mapper[node_id] for node_id in eligibles])
        weights = weights.max() - weights + 1
        weights = weights / weights.sum()

        selected_id = rand.choices(eligibles, weights=weights)[0]

        ordered_node_ids.append(selected_id)
        selected_ids_set.add(selected_id)
        candidates.remove(selected_id)
        candidates.update([child_id for child_id in node_id2child_ids[selected_id]])

    ordered_nodes = sorted(head_nodes, key=lambda node: ordered_node_ids.index(node.id))

    return ordered_nodes



from sampo.base import SAMPO
from sampo.schemas import WorkGraph, GraphNode


def update_priority(node, priority_value, visited=None):
    if visited is None:
        visited = set()

    visited.add(node)

    node.work_unit.priority = min(node.work_unit.priority, priority_value)

    for parent in getattr(node, "parents", []):
        update_priority(parent, priority_value, visited)


def check_and_correct_priorities(wg: WorkGraph, verbose: bool = True):
    # check priorities
    for node in wg.nodes:
        # if node.is_inseparable_son():
        #     continue
        for parent_node in node.parents:
            if node.work_unit.priority < parent_node.work_unit.priority:
                if verbose:
                    SAMPO.logger.warn(f'Priority of node {node.work_unit.display_name} '
                                      f'is less than node {parent_node.work_unit.name} '
                                      f'conflicting with edges, applying auto-correction')
                update_priority(parent_node, node.work_unit.priority)


def extract_priority_groups_from_nodes(nodes: list[GraphNode]) -> dict[int, list[GraphNode]]:
    groups = {priority: [] for priority in set(node.work_unit.priority for node in nodes)}
    for node in nodes:
        groups[node.work_unit.priority].append(node)
    return groups


def extract_priority_groups_from_indices(priorities: list[int]) -> dict[int, list[int]]:
    groups = {priority: [] for priority in set(priorities)}
    for node, priority in enumerate(priorities):
        groups[priority].append(node)
    return groups



from operator import itemgetter

from sampo.utilities.priority import extract_priority_groups_from_nodes
from sampo.scheduler.utils.time_computaion import work_priority, calculate_working_time_cascade
from sampo.schemas.graph import GraphNode
from sampo.schemas.time_estimator import WorkTimeEstimator


def ford_bellman(nodes: list[GraphNode],
                 weights: dict[str, float],
                 node_id2parent_ids: dict[str, set[str]]) -> dict[str, float]:
    """
    Runs heuristic ford-bellman algorithm for given graph and weights.
    """
    path_weights: dict[str, float] = {node.id: 0 for node in nodes}
    # cache graph edges
    edges: list[tuple[str, str, float]] = sorted([(finish, start.id, weights[finish])
                                                  for start in nodes
                                                  for finish in node_id2parent_ids[start.id]
                                                  if finish in path_weights])
    if not edges:
        return path_weights

    # for changes heuristic
    changed = False
    # run standard ford-bellman on reversed edges
    # optimize dict access to finish weight
    for i in range(len(nodes)):
        cur_finish = edges[0][0]
        cur_finish_weight = path_weights[cur_finish]
        # relax on edges
        for finish, start, weight in edges:
            # we are running through the equality class by finish node
            # so if it changes renew the parameters of current equality class
            if cur_finish != finish:
                path_weights[cur_finish] = cur_finish_weight
                cur_finish = finish
                cur_finish_weight = path_weights[cur_finish]
            new_weight = path_weights[start] + weight
            if new_weight < cur_finish_weight:
                cur_finish_weight = new_weight
                changed = True
        # if we were done completely nothing with actual minimum weights, the algorithm ends
        if not changed:
            break
        # we go here if changed = True
        # so the last equality class weight can be changed, save it
        path_weights[cur_finish] = cur_finish_weight
        # next iteration should start without change info from previous
        changed = False

    return path_weights


def prioritization_nodes(nodes: list[GraphNode],
                         node_id2parent_ids: dict[str, set[str]],
                         work_estimator: WorkTimeEstimator) -> list[GraphNode]:
    """
    Return ordered critical nodes.
    Finish time is depended on these ordered nodes.
    """
    if len(nodes) == 1:
        return nodes

    # inverse weights
    weights = {node.id: -work_priority(node, calculate_working_time_cascade, work_estimator)
               for node in nodes}

    path_weights = ford_bellman(nodes, weights, node_id2parent_ids)

    ordered_nodes = sorted(nodes, key=lambda node: path_weights[node.id])

    return ordered_nodes


def prioritization(head_nodes: list[GraphNode],
                   node_id2parent_ids: dict[str, set[str]],
                   node_id2child_ids: dict[str, set[str]],
                   work_estimator: WorkTimeEstimator) -> list[GraphNode]:
    """
    Return ordered critical nodes.
    Finish time is depended on these ordered nodes.
    """

    # form groups
    groups = extract_priority_groups_from_nodes(head_nodes)

    result = []
    for _, group in sorted(groups.items(), key=itemgetter(0)):
        result.extend(prioritization_nodes(group, node_id2parent_ids, work_estimator))

    return result



def intensity(vehicle_num: int, length: float, max_velocity: float):
    return vehicle_num * max_velocity / length


def static_workload(vehicle_num: int, length: float, max_velocity: float, bandwidth: float):
    """
    Calculate rate of theoretical road workload
    :param bandwidth:
    :param vehicle_num:
    :param length:
    :param max_velocity:
    :return:
    """
    return intensity(vehicle_num, length, max_velocity) / bandwidth



import uuid
from dataclasses import dataclass
from functools import cached_property
from typing import Any

import numpy as np

from sampo.schemas.exceptions import NoAvailableResources


@dataclass
class LandEdge:
    """
    A representation of the connection between two vertices of a transport graph
    (e.x., the edge between platform and warehouse)
    """
    id: str
    start: 'LandGraphNode'
    finish: 'LandGraphNode'
    weight: float
    bandwidth: int


class ResourceStorageUnit:
    def __init__(self, capacity: dict[str, int] | None = None):
        """
        Represents the resource storage of a land graph node
        :param capacity: list of the maximum values of each material
        """
        if capacity is None:
            capacity = {}
        self._capacity = capacity

    @cached_property
    def capacity(self) -> dict[str, int]:
        return self._capacity


class LandGraphNode:
    def __init__(self,
                 id: str = str(uuid.uuid4()),
                 name: str = 'platform',
                 resource_storage_unit: ResourceStorageUnit = ResourceStorageUnit(),
                 neighbour_nodes: list[tuple['LandGraphNode', float, int]] | None = None,
                 works: list['GraphNode'] | Any = None):
        """
        Represents the participant of landscape transport network

        :param neighbour_nodes: parent nodes list that saves parent itself and the weight of edge
        :param resource_storage_unit: object that saves info about the resources in the current node
        (in other words platform)
        """
        self.id = id
        self.name = name
        if not (neighbour_nodes is None):
            self.add_neighbours(neighbour_nodes)
        self._roads: list[LandEdge] = []
        self.nodes: list['GraphNode'] = []
        if works is not None:
            self.nodes.extend(works) if isinstance(works, list) else self.nodes.append(works)
        self.resource_storage_unit = resource_storage_unit

    def __hash__(self):
        return hash(self.id)

    @cached_property
    def neighbours(self) -> list['LandGraphNode']:
        if self._roads:
            return [neighbour_edge.finish for neighbour_edge in self._roads]
        return []

    @cached_property
    def roads(self) -> list[LandEdge]:
        if self._roads:
            return self._roads
        raise NoAvailableResources('There are no roads in land graph')

    def add_works(self, nodes: list['GraphNode'] | Any):
        if isinstance(nodes, list):
            self.nodes.extend(nodes)
        else:
            self.nodes.append(nodes)

    @cached_property
    def works(self) -> list['GraphNode']:
        return self.nodes

    def add_neighbours(self, neighbour_nodes: list[tuple['LandGraphNode', float, int]] | tuple['LandGraphNode', float, int]):
        if isinstance(neighbour_nodes, list):
            for neighbour, length, bandwidth in neighbour_nodes:
                road_id = str(uuid.uuid4())
                self._roads.append(LandEdge(road_id, self, neighbour, length, bandwidth))
                neighbour._roads.append(LandEdge(road_id, neighbour, self, length, bandwidth))
        else:
            neighbour, length, bandwidth = neighbour_nodes
            road_id = str(uuid.uuid4())
            self._roads.append(LandEdge(road_id, self, neighbour, length, bandwidth))
            neighbour._roads.append(LandEdge(road_id, neighbour, self, length, bandwidth))

@dataclass
class LandGraph:
    nodes: list[LandGraphNode] = None
    adj_matrix: np.array = None
    node2ind: dict[LandGraphNode, int] = None
    id2ind: dict[str, int] = None
    vertex_count: int = None

    def __post_init__(self) -> None:
        self.reinit()

    def reinit(self):
        adj_matrix, node2ind, id2ind = self._to_adj_matrix()
        object.__setattr__(self, 'adj_matrix', adj_matrix)
        object.__setattr__(self, 'node2ind', node2ind)
        object.__setattr__(self, 'id2ind', id2ind)
        object.__setattr__(self, 'vertex_count', len(node2ind))

    @cached_property
    def edges(self) -> list['LandEdge']:
        # TODO: to do
        road_ids = set()
        roads = []

        for node in self.nodes:
            for road in node.roads:
                if road.id not in road_ids:
                    road_ids.add(road.id)
                    roads.append(road)

        return roads

    def _to_adj_matrix(self) -> tuple[np.array, dict[LandGraphNode, int], dict[str, int]]:
        node2ind: dict[LandGraphNode, int] = {
            v: i for i, v in enumerate(self.nodes)
        }
        id2ind = {
            v.id: i for i, v in enumerate(self.nodes)
        }
        adj_mtrx = np.full((len(node2ind), len(node2ind)), np.inf)
        for v, i in node2ind.items():
            for child in v.roads:
                c_i = node2ind[child.finish]
                adj_mtrx[i, c_i] = child.weight

        return adj_mtrx, node2ind, id2ind



from random import Random
from typing import Callable

from sampo.generator.config.gen_counts import MIN_GRAPH_COUNTS, ADDITION_CLUSTER_PROBABILITY, GRAPH_COUNTS, \
    MAX_BOREHOLES_PER_BLOCK, BRANCHING_PROBABILITY
from sampo.generator.pipeline import SyntheticGraphType, StageType
from sampo.generator.pipeline.cluster import get_cluster_works, _add_addition_work
from sampo.schemas.graph import GraphNode, WorkGraph


def get_small_graph(cluster_name: str | None = 'C1', rand: Random | None = None) -> WorkGraph:
    """
    Creates a small graph of works consisting of 30-50 vertices

    :param cluster_name: str - the first cluster name
    :param rand: Optional[Random] - generator of numbers with a given seed or None
    :return: work_graph: WorkGraph - work graph where count of vertex between 30 and 50
    """

    pipe_nodes_count = MIN_GRAPH_COUNTS['pipe_nodes'].rand_int(rand)
    pipe_net_count = MIN_GRAPH_COUNTS['pipe_net'].rand_int(rand)
    light_masts_count = MIN_GRAPH_COUNTS['light_masts'].rand_int(rand)
    borehole_count = MIN_GRAPH_COUNTS['borehole'].rand_int(rand)

    c1, c_nodes, _ = get_cluster_works(cluster_name=cluster_name, pipe_nodes_count=pipe_nodes_count,
                                       pipe_net_count=pipe_net_count, light_masts_count=light_masts_count,
                                       borehole_counts=[borehole_count], rand=rand)
    return WorkGraph.from_nodes(list(c_nodes.values()) + [c1])


def _get_cluster_graph(cluster_name: str, pipe_nodes_count: int | None = None,
                       pipe_net_count: int | None = None, light_masts_count: int | None = None,
                       borehole_counts: list[int] | None = None, add_addition_cluster: bool | None = False,
                       addition_cluster_probability: float | None = ADDITION_CLUSTER_PROBABILITY,
                       rand: Random | None = None) -> tuple[list[GraphNode], dict[str, GraphNode], int]:
    pipe_nodes_count = pipe_nodes_count or GRAPH_COUNTS['pipe_nodes'].rand_int(rand)
    pipe_net_count = pipe_net_count or GRAPH_COUNTS['pipe_net'].rand_int(rand)
    light_masts_count = light_masts_count or GRAPH_COUNTS['light_masts'].rand_int(rand)
    if borehole_counts is None:
        counts = GRAPH_COUNTS['borehole'].rand_int(rand)
        if counts <= MAX_BOREHOLES_PER_BLOCK:
            borehole_counts = [counts]
        else:
            borehole_counts = [MAX_BOREHOLES_PER_BLOCK, counts - MAX_BOREHOLES_PER_BLOCK]

    # the whole count of nodes generated in this function
    count_nodes = 0
    c_master, roads, count_master = get_cluster_works(cluster_name=cluster_name,
                                                      pipe_nodes_count=pipe_nodes_count, pipe_net_count=pipe_net_count,
                                                      light_masts_count=light_masts_count,
                                                      borehole_counts=borehole_counts, rand=rand)
    count_nodes += count_master

    checkpoints = [c_master]
    if add_addition_cluster or _add_addition_work(addition_cluster_probability, rand):
        c_slave, _, count_slave = get_cluster_works(cluster_name=cluster_name,
                                                    pipe_nodes_count=pipe_nodes_count, pipe_net_count=pipe_net_count,
                                                    light_masts_count=light_masts_count, borehole_counts=borehole_counts,
                                                    roads=roads, rand=rand)
        count_nodes += count_slave
        checkpoints.append(c_slave)
    return checkpoints, roads, count_nodes


def get_graph(mode: SyntheticGraphType | None = SyntheticGraphType.GENERAL,
              cluster_name_prefix: str | None = 'C',
              cluster_counts: int | None = 0,
              branching_probability: float | None = BRANCHING_PROBABILITY,
              addition_cluster_probability: float | None = ADDITION_CLUSTER_PROBABILITY,
              bottom_border: int | None = 0,
              top_border: int | None = 0,
              rand: Random | None = None) -> WorkGraph:
    """
    Invokes a graph of the given type if at least one positive value of cluster_counts, addition_cluster_probability,
    bottom_border is given

    :param mode: str - 'general' or 'sequence' or 'parallel - the type of the returned graph
    :param cluster_name_prefix: str -  cluster name prefix, if the prefix is 'C',
        the clusters will be called 'C1', 'C2' etc.
    :param cluster_counts: Optional[int] - Number of clusters for the graph
    :param branching_probability: Optional[float] - The probability that the node will not be connected to the last
        cluster, but to any other cluster for the general mode
    :param addition_cluster_probability: Optional[float] - probability of a slave (example C3_1) cluster
        to the main cluster
    :param bottom_border: Optional[int] - bottom border for number of works for the graph
    :param top_border: Optional[int] - top border for number of works for the graph
    :param rand: Optional[Random] - generator of numbers with a given seed or None
    :return:
        work_graph: WorkGraph - the desired work graph
    """
    assert cluster_counts + bottom_border + top_border > 0, 'At least one border param should be specified'
    assert cluster_counts >= 0 and branching_probability >= 0 and top_border >= 0, 'Params should not be negative'

    rand = rand or Random()
    get_root_stage: Callable[[list[StageType], float, Random], GraphNode] = _graph_mode_to_callable(mode)

    if bottom_border > 0:
        top_border = 0
    masters_clusters_ind = 1
    works_generated = 0
    stages = []

    while True:
        if cluster_counts > 0 and cluster_counts == len(stages) - 1:
            addition_cluster_probability = 0

        checkpoints, roads, count_works = _get_cluster_graph(f'{cluster_name_prefix}{masters_clusters_ind}',
                                                             addition_cluster_probability=addition_cluster_probability,
                                                             rand=rand)
        root_stage = get_root_stage(stages, branching_probability, rand)
        if root_stage is not None:
            for checkpoint in checkpoints:
                checkpoint.add_parents([root_stage])

        stages += [(c, roads) for c in checkpoints]
        masters_clusters_ind += 1
        works_generated += count_works

        if (0 < bottom_border <= works_generated or 0 < top_border < count_works + works_generated
                or 0 < cluster_counts <= (len(stages) - 1)):
            break

    if len(stages) == 1:
        return get_small_graph(cluster_name=f'{cluster_name_prefix}1')

    nodes = [road for _, roads in stages for road in roads.values()]
    nodes.extend([c for c, _ in stages])
    return WorkGraph.from_nodes(nodes)


def _graph_mode_to_callable(mode: SyntheticGraphType) -> \
        Callable[[list[tuple[GraphNode, dict[str, GraphNode]]], float, Random], GraphNode]:
    if mode is SyntheticGraphType.GENERAL:
        return _general_graph_mode
    if mode is SyntheticGraphType.PARALLEL:
        return _parallel_graph_mode_get_root
    return _sequence_graph_mode_get_root


def _parallel_graph_mode_get_root(stages: list[tuple[GraphNode, dict[str, GraphNode]]],
                                  branching_probability: float, rand: Random) -> GraphNode | None:
    if len(stages) == 0:
        return None
    return stages[0][0]


def _sequence_graph_mode_get_root(stages: list[tuple[GraphNode, dict[str, GraphNode]]],
                                  branching_probability: float, rand: Random) -> GraphNode | None:
    if len(stages) == 0:
        return None
    return stages[-1][0]


def _general_graph_mode(stages: list[tuple[GraphNode, dict[str, GraphNode]]],
                        branching_probability: float, rand: Random) -> GraphNode | None:
    is_branching = rand.random() <= branching_probability
    ind = len(stages) - 1
    if is_branching and len(stages) > 2:
        ind = rand.randint(1, len(stages) - 2)
    if len(stages) == 0:
        return None
    return stages[ind][1]['min']



from sampo.schemas.graph import GraphNode
from sampo.schemas.time import Time


def get_exec_times_from_assigned_time_for_chain(inseparable_chain: list[GraphNode],
                                                assigned_time: Time) -> dict[GraphNode, tuple[Time, Time]]:
    """
    Distributes a given total execution time among work nodes in an inseparable chain.

    The time distribution is proportional to each node's volume, ensuring that
    the entire `assigned_time` is utilized. Any rounding discrepancies are
    allocated to the last node in the chain.

    Args:
        inseparable_chain: A list of nodes representing an inseparable sequence of work units.
        assigned_time: The total `Time` allocated for the entire chain's execution.

    Returns:
        A dictionary mapping each `GraphNode` to a tuple `(lag, node_execution_time)`.
        `lag` is always `Time(0)` as the chain is inseparable, and
        `node_execution_time` is the calculated execution time for that specific node.
    """
    total_volume = sum(n.work_unit.volume for n in inseparable_chain)

    # Handle the edge case where total_volume is zero or negative.
    # In this scenario, time is distributed equally among nodes,
    # with any remainder allocated to the last node.
    if total_volume <= 0:
        node_time = assigned_time // len(inseparable_chain)
        # Distribute time equally among all but the last node
        exec_times = {n: (Time(0), node_time) for n in inseparable_chain[:-1]}
        # Assign remaining time to the last node
        exec_times[inseparable_chain[-1]] = Time(0), assigned_time - node_time * len(exec_times)
        return exec_times

    exec_times: dict[GraphNode, tuple[Time, Time]] = {}
    remaining_time = assigned_time  # Initialize remaining time to distribute

    # Iterate through all nodes except the last one.
    # Time is distributed based on the node's volume relative to the *remaining* total volume.
    for i, n in enumerate(inseparable_chain[:-1]):
        # Calculate the proportion of the current node's volume to the remaining total volume.
        volume_proportion = 0
        if total_volume > 0:
            volume_proportion = n.work_unit.volume / total_volume

        # Calculate execution time for the current node and convert to integer.
        # This takes a portion of the *remaining* time.
        exec_time = int(remaining_time.value * volume_proportion)
        exec_times[n] = (Time(0), Time(exec_time))

        # Deduct the current node's volume and allocated time from the totals
        total_volume -= n.work_unit.volume
        remaining_time -= exec_time

    # The last node receives all remaining time to account for any rounding errors
    # during the distribution to previous nodes.
    exec_times[inseparable_chain[-1]] = (Time(0), remaining_time)
    return exec_times



import random
from typing import Optional

from sampo.schemas.utils import uuid_str
from sampo.schemas.works import WorkUnit
from sampo.utilities.sampler.requirements import get_worker_reqs_list
from sampo.utilities.sampler.types import MinMax


def get_work_unit(rand: random.Random, name: str,
                  work_id: Optional[str] = '',
                  volume_type: Optional[str] = 'unit',
                  group: Optional[str] = 'default',
                  work_volume: Optional[MinMax[float]] = MinMax[float](0.1, 100.0),
                  req_volume: Optional[MinMax[int]] = MinMax[int](1, 50),
                  req_worker_count: Optional[MinMax[int]] = MinMax[int](1, 100)
                  ) -> WorkUnit:
    reqs = get_worker_reqs_list(rand, req_volume, req_worker_count)
    work_id = work_id or uuid_str(rand)
    volume = rand.random() * (work_volume.max - work_volume.min) + work_volume.min
    return WorkUnit(work_id, name, worker_reqs=reqs, volume=volume, volume_type=volume_type, group=group)


def get_similar_work_unit(rand: random.Random,
                          exemplar: WorkUnit,
                          scalar: Optional[float] = 1.0,
                          name: Optional[str] = '',
                          work_id: Optional[str] = ''
                          ) -> WorkUnit:
    reqs = [req.scale_all(scalar) for req in exemplar.worker_reqs]
    work_id = work_id or uuid_str(rand)
    name = name or exemplar.name
    return WorkUnit(work_id, name, worker_reqs=reqs, group=exemplar.group,
                    volume=exemplar.volume * scalar, volume_type=exemplar.volume_type)



from abc import ABC, abstractmethod
from enum import Enum
from functools import partial
from typing import Callable

import numpy as np
from deap import base, creator

from sampo.schemas import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec

ChromosomeType = tuple[np.ndarray, np.ndarray, np.ndarray, ScheduleSpec, np.ndarray]

class ScheduleGenerationScheme(Enum):
    Parallel = 'Parallel'
    Serial = 'Serial'


class FitnessFunction(ABC):
    """
    Base class for description of different fitness functions.
    """

    @abstractmethod
    def evaluate(self, chromosome: ChromosomeType, evaluator: Callable[[ChromosomeType], Schedule]) \
            -> tuple[int | float]:
        """
        Calculate the value of fitness function of the chromosome.
        It is better when value is less.
        """
        ...


# create class FitnessMin, the weights = -1 means that fitness - is function for minimum

# creator.create('FitnessMin', base.Fitness, weights=(-1.0,))
# creator.create('Individual', list, fitness=creator.FitnessMin)
# Individual = creator.Individual

class Individual(list):
    def __init__(self, individual_fitness_constructor: Callable[[], base.Fitness], chromosome: ChromosomeType):
        super().__init__(chromosome)
        self.fitness = individual_fitness_constructor()

    @staticmethod
    def prepare(individual_fitness_constructor: Callable[[], base.Fitness]) -> Callable[[ChromosomeType], list]:
        """
        Returns the constructor of Individual prepared to use in Genetic algorithm
        """
        return partial(Individual, individual_fitness_constructor)




from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional
from uuid import uuid4

from sampo.schemas.resources import Material
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.schemas.time import Time
from sampo.schemas.zones import Zone

# Used for max_count in the demand, if it is not specified during initialization WorkerReq
DEFAULT_MAX_COUNT = 100


class BaseReq(AutoJSONSerializable['BaseReq'], ABC):
    """
    A class summarizing any requirements for the work to be performed related to renewable and non-renewable
    resources, infrastructure requirements, etc.
    """
    @property
    @abstractmethod
    def name(self) -> str:
        """
        Returns the name of the claim if it exists, e.g. 'dig work claim'.

        :return name: the name of req
        """
        ...


@dataclass
class WorkerReq(BaseReq):
    """
    Requirements related to renewable human resources

    :param kind: type of resource/profession
    :param volume: volume of work in time units
    :param min_count: minimum number of employees needed to perform the work
    :param max_count: maximum allowable number of employees performing the work
    :param name: the name of this requirement
    """

    def __init__(self,
                 kind: str,
                 volume: Time,
                 min_count: Optional[int] = 1,
                 max_count: Optional[int] = DEFAULT_MAX_COUNT,
                 name: Optional[str] = ''):
        self.kind = kind
        self.volume = volume
        self.min_count = min_count
        self.max_count = max_count
        self.name = name

        assert min_count <= max_count

    def name(self):
        return self.name


    def scale_all(self, scalar: float, new_name: Optional[str] = '') -> 'WorkerReq':
        """
        The function scales the requirement to the size of the work including the total
        volume and the maximum number of personnel involved.

        :param scalar: scalar for multiplication
        :param new_name: name for new req
        :return new_req: new object with new volume of the work and extended max_count_commands
        """
        max_count = max(round(self.max_count * scalar), self.min_count)
        new_req = WorkerReq(self.kind, self.volume * scalar, self.min_count, max_count, new_name or self.name)
        return new_req

    def scale_volume(self, scalar: float, new_name: Optional[str] = None) -> 'WorkerReq':
        """
        The function scales only volume of the work for the requirement.

        :param scalar: scalar for multiplication
        :param new_name: name for new req
        :return new_req: new object with new volume of the work.
        """
        new_req = WorkerReq(self.kind, self.volume * scalar, self.min_count, self.max_count, new_name or self.name)
        return new_req


@dataclass(frozen=True)
class EquipmentReq(BaseReq):
    """
    Requirements for renewable non-human resources: equipment, trucks, machines, etc

    :param kind: type of resource/profession
    :param name: the name of this requirement
    """
    kind: str
    count: int
    name: Optional[str] = None


@dataclass(frozen=True)
class MaterialReq(BaseReq):
    """
    Requirements for non-renewable materials: consumables, spare parts, construction materials

    :param kind: type of resource/profession
    :param name: the name of this requirement
    """
    kind: str
    count: int
    # need_start: int TODO Implement handling this
    name: Optional[str] = None

    def material(self) -> Material:
        return Material(str(uuid4()), self.kind, self.count)


@dataclass(frozen=True)
class ConstructionObjectReq(BaseReq):
    """
    Requirements for infrastructure and the construction of other facilities: electricity, pipelines, roads, etc

    :param kind: type of resource/profession
    :param name: the name of this requirement
    """
    kind: str
    count: int
    name: Optional[str] = None


@dataclass(frozen=True)
class ZoneReq(BaseReq):
    kind: str
    required_status: int
    name: Optional[str] = None

    def to_zone(self) -> Zone:
        return Zone(self.kind, self.required_status)



from abc import ABC, abstractmethod

from sampo.api.genetic_api import ChromosomeType, Individual
from sampo.scheduler import GeneticScheduler, Scheduler
from sampo.scheduler.genetic.schedule_builder import create_toolbox
from sampo.schemas import WorkGraph, Contractor, Time, LandscapeConfiguration
from sampo.schemas.schedule_spec import ScheduleSpec


class PopulationScheduler(ABC):

    @abstractmethod
    def schedule(self,
                 initial_population: list[ChromosomeType],
                 wg: WorkGraph,
                 contractors: list[Contractor],
                 spec: ScheduleSpec = ScheduleSpec(),
                 assigned_parent_time: Time = Time(0),
                 landscape: LandscapeConfiguration = LandscapeConfiguration()) -> list[ChromosomeType]:
        ...


class GeneticPopulationScheduler(PopulationScheduler):
    def __init__(self, genetic: GeneticScheduler):
        self._genetic = genetic

    def schedule(self,
                 initial_population: list[ChromosomeType],
                 wg: WorkGraph,
                 contractors: list[Contractor],
                 spec: ScheduleSpec = ScheduleSpec(),
                 assigned_parent_time: Time = Time(0),
                 landscape: LandscapeConfiguration = LandscapeConfiguration()) -> list[ChromosomeType]:
        return self._genetic.upgrade_pop(wg, contractors, initial_population, spec,
                                         assigned_parent_time, landscape=landscape)


class HeuristicPopulationScheduler(PopulationScheduler):
    def __init__(self, schedulers: list[Scheduler]):
        self._schedulers = schedulers

    def schedule(self,
                 initial_population: list[ChromosomeType],
                 wg: WorkGraph,
                 contractors: list[Contractor],
                 spec: ScheduleSpec = ScheduleSpec(),
                 assigned_parent_time: Time = Time(0),
                 landscape: LandscapeConfiguration = LandscapeConfiguration()) -> list[ChromosomeType]:
        toolbox = create_toolbox(wg=wg, contractors=contractors,
                                 spec=spec, assigned_parent_time=assigned_parent_time,
                                 landscape=landscape)
        return [toolbox.Individual(toolbox.schedule_to_chromosome(schedule=schedule))
                for scheduler in self._schedulers
                for schedule in scheduler.schedule(wg, contractors, spec, landscape=landscape)]
